/*
 * Copyright (c) 2025 Alexander Bokovoy <ab@samba.org>
 * Copyright (c) 2025 Andreas Schneider <asn@samba.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define _GNU_SOURCE 1

#include <errno.h>
#include <arpa/inet.h>
#include <kdb.h>
#include <stdio.h>
#include <sys/types.h>
#include <ifaddrs.h>
#include <net/if.h>
#include <netdb.h>
#include <profile.h>

#include <kirmes/kirmes.h>
#include <krb5/krb5.h>
#include <string.h>

struct kuserdb_context {
    krb5_context kdb_base_ctx;
    bool disallow_name_aliases;
};

struct ksock_address {
	socklen_t sa_socklen;
	union {
		struct sockaddr s;
		struct sockaddr_in in;
		struct sockaddr_in6 in6;
		struct sockaddr_storage ss;
	} sa;
};

static int parse_ip_addr(const char *hostaddr, struct ksock_address *kaddr)
{
    int rc;

    rc = inet_pton(AF_INET6, hostaddr, &kaddr->sa.in6.sin6_addr);
    if (rc == 1) {
        kaddr->sa.in6.sin6_family = AF_INET6;
        kaddr->sa_socklen = sizeof(struct sockaddr_in6);
        return 0;
    }

    rc = inet_pton(AF_INET, hostaddr, &kaddr->sa.in.sin_addr);
    if (rc == 1) {
        kaddr->sa.in.sin_family = AF_INET;
        kaddr->sa_socklen = sizeof(struct sockaddr_in);
        return 0;
    }

    return -1;
}

static krb5_error_code
kuserdb_lib_init(void)
{
    return 0;
}

static krb5_error_code
kuserdb_lib_cleanup(void)
{
    return 0;
}

static krb5_error_code
kuserdb_create_context(krb5_context context, const char *conf_section)
{
    struct kuserdb_context *kctx = NULL;
    struct _profile_t *profile = NULL;
    krb5_error_code ret;
    int val = 0;

    kctx = calloc(1, sizeof(struct kuserdb_context));
    if (kctx == NULL) {
        return ENOMEM;
    }

    ret = krb5_get_profile(context, &profile);
    if (ret != 0) {
        goto cleanup;
    }

    ret = profile_get_boolean(profile,
                              KDB_MODULE_SECTION,
                              conf_section,
                              "disallow_name_aliases",
                              true,
                              &val);
    profile_release(profile);
    if (ret != 0) {
        goto cleanup;
    }
    kctx->disallow_name_aliases = val;

    ret = krb5_copy_context(context, &kctx->kdb_base_ctx);
    if (ret != 0) {
        goto cleanup;
    }

    ret = krb5_db_load_module(kctx->kdb_base_ctx, "klmdb");
    if (ret != 0) {
        goto cleanup;
    }

    ret = krb5_db_set_context(context, kctx);
    if (ret != 0) {
        goto cleanup;
    }

cleanup:
    if (ret != 0) {
        free(kctx);
    }

    return ret;
}

static void
kuserdb_destroy_context(struct kuserdb_context *ctx)
{
    krb5_free_context(ctx->kdb_base_ctx);
    free(ctx);
}

static krb5_error_code
kuserdb_open(krb5_context context,
             char *conf_section,
             char **db_args,
             int mode)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret == 0) {
        goto open;
    }

    ret = kuserdb_create_context(context, conf_section);
    if (ret != 0) {
        return ret;
    }

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

open:
    return krb5_db_open(kctx->kdb_base_ctx, db_args, mode);
}

static krb5_error_code
kuserdb_fini(krb5_context context)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_fini(kctx->kdb_base_ctx);
}

static krb5_error_code
kuserdb_create(krb5_context context,
               char *conf_section,
               char **db_args)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret == 0) {
        goto create;
    }

    ret = kuserdb_create_context(context, conf_section);
    if (ret != 0) {
        return ret;
    }

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

create:
    return krb5_db_create(kctx->kdb_base_ctx, db_args);
}

static krb5_error_code
kuserdb_destroy(krb5_context context,
                char *unused, /* conf_section */
                char **db_args)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    ret = krb5_db_destroy(kctx->kdb_base_ctx, db_args);

    kuserdb_destroy_context(kctx);
    krb5_db_set_context(context, NULL);

    return ret;
}

static krb5_error_code
update_db_entry_from_user_record(krb5_context context,
                                 const struct KirmesUserRecord *record,
                                 krb5_db_entry *entry)
{
    krb5_error_code ret;

    krb5_free_principal(context, entry->princ);

    /* Update the db_entry with information from systemd's userdb */
    ret = krb5_parse_name(context, record->user_name, &entry->princ);
    if (ret != 0) {
        return ret;
    }

    /* The entries we create for userdb will have set KRB5_KDB_DISALLOW_ALL_TIX
     * by default!
     */
    entry->attributes &= ~KRB5_KDB_DISALLOW_ALL_TIX;

    return 0;
}

static krb5_error_code
match_address(krb5_context context,
              struct kuserdb_context *kctx,
              krb5_const_principal searchfor,
              unsigned int flags,
              struct ifaddrs *iflist,
              struct ksock_address *kaddr,
              krb5_db_entry **entry_out)
{
    krb5_principal hostname_principal = NULL;
    struct ifaddrs *ifptr = NULL;
    char ip_hostname[NI_MAXHOST] = {0};
    krb5_error_code ret;
    int cmp = -1;
    int rc;

    for (ifptr = iflist; ifptr != NULL; ifptr = ifptr->ifa_next) {
        if (ifptr->ifa_addr == NULL || ifptr->ifa_netmask == NULL) {
            continue;
        }
        if ((ifptr->ifa_flags & IFF_UP) == 0) {
            continue;
        }

        if (ifptr->ifa_addr->sa_family != kaddr->sa.ss.ss_family) {
            continue;
        }

        if (ifptr->ifa_addr->sa_family == AF_INET6) {
            struct sockaddr_in6 *sin6 =
                (struct sockaddr_in6 *)(uintptr_t)ifptr->ifa_addr;

            cmp = memcmp(sin6, &kaddr->sa.in6, kaddr->sa_socklen);
        } else if (ifptr->ifa_addr->sa_family == AF_INET) {
            struct sockaddr_in *sin =
                (struct sockaddr_in *)(uintptr_t)ifptr->ifa_addr;

            cmp = memcmp(sin, &kaddr->sa.in, kaddr->sa_socklen);
        }
        if (cmp == 0) {
            break;
        }
    }

    if (cmp != 0) {
        return KRB5_KDB_NOENTRY;
    }

    rc = getnameinfo(&kaddr->sa.s,
                     kaddr->sa_socklen,
                     ip_hostname,
                     sizeof(ip_hostname),
                     NULL,
                     0,
                     NI_NAMEREQD);
    if (rc != 0) {
        return KRB5_KDB_NOENTRY;
    }

    ret = krb5_copy_principal(context, searchfor, &hostname_principal);
    if (ret != 0) {
        return ret;
    }

    krb5_data data = {.length = strlen(ip_hostname), .data = ip_hostname};

    krb5_data old_data = *krb5_princ_component(context, hostname_principal, 1);
    krb5_data *data_ptr = krb5_princ_component(context, hostname_principal, 1);

    data_ptr->length = data.length;
    data_ptr->data = data.data;


    ret = krb5_db_get_principal(kctx->kdb_base_ctx,
                                hostname_principal,
                                flags,
                                entry_out);


    data_ptr->length = old_data.length;
    data_ptr->data = old_data.data;
    krb5_free_principal(context, hostname_principal);

    return ret;
}

static krb5_error_code
kuserdb_get_service_alias(krb5_context context,
                          struct kuserdb_context *kctx,
                          krb5_const_principal searchfor,
                          unsigned int flags,
                          krb5_db_entry **entry_out)
{
    struct ksock_address kaddr = {
        .sa_socklen = sizeof(struct sockaddr_storage),
    };
    struct ifaddrs *iflist = NULL;
    struct addrinfo *ai = NULL;
    krb5_data *host = NULL;
    krb5_error_code ret;
    int rc;

    if (krb5_princ_size(context, searchfor) > 2) {
        return KRB5_KDB_NOENTRY;
    }

    host = krb5_princ_component(context, searchfor, 1);
    if (host->length == 0) {
        return KRB5_KDB_NOENTRY;
    }

    rc = getifaddrs(&iflist);
    if (rc != 0) {
        return KRB5_KDB_NOENTRY;
    }

    {
        char host_or_addr[host->length + 1];
        memcpy(host_or_addr, host->data, host->length);
        host_or_addr[host->length] = '\0';

        rc = parse_ip_addr(host_or_addr, &kaddr);
        if (rc == 0) {
            ret = match_address(context,
                                kctx,
                                searchfor,
                                flags,
                                iflist,
                                &kaddr,
                                entry_out);
            goto out;
        }

        /* This could be used for DoS */
        if (!kctx->disallow_name_aliases) {
            struct addrinfo *ai_ptr = NULL;

            rc = getaddrinfo(host_or_addr, NULL, NULL, &ai);
            if (rc != 0) {
                ret = KRB5_KDB_NOENTRY;
                goto out;
            }

            for (ai_ptr = ai; ai_ptr != NULL; ai_ptr = ai_ptr->ai_next) {
                kaddr.sa_socklen = ai_ptr->ai_addrlen;
                memcpy(&kaddr.sa.s, ai_ptr->ai_addr, ai_ptr->ai_addrlen);

                ret = match_address(context,
                                    kctx,
                                    searchfor,
                                    flags,
                                    iflist,
                                    &kaddr,
                                    entry_out);
                if (ret == 0) {
                    goto out;
                }
            }
        }
    }

    ret = KRB5_KDB_NOENTRY;
out:
    freeaddrinfo(ai);
    freeifaddrs(iflist);
    return ret;
}

static krb5_error_code
kuserdb_get_userdb_principal(krb5_context context,
                             struct kuserdb_context *kctx,
                             krb5_const_principal searchfor,
                             unsigned int flags,
                             krb5_db_entry **entry_out)
{
    struct KirmesUserRecord *records = NULL;
    char userdb_name[256] = {0};
    krb5_principal userdb_principal = NULL;
    size_t num_records = 0;
    enum KirmesStatus status;
    char *user = NULL;
    krb5_error_code ret;
    int rc;

    ret = krb5_unparse_name_flags(context,
                                  searchfor,
                                  KRB5_PRINCIPAL_UNPARSE_SHORT,
                                  &user);
    if (ret != 0) {
        return KRB5_KDB_NOENTRY;
    }

    status = kirmes_get_user_record(NULL,
                                    user,
                                    "io.systemd.Multiplexer",
                                    &records,
                                    &num_records);
    free(user);
    if (status != KirmesStatusSuccess) {
        return KRB5_KDB_NOENTRY;
    }

    rc = snprintf(userdb_name,
                  sizeof(userdb_name),
                  "userdb:%s",
                  records[0].user_name);
    if (rc < 0) {
        kirmes_user_record_free(records);
        return KRB5_KDB_NOENTRY;
    }
    ret = krb5_parse_name(context, userdb_name, &userdb_principal);
    if (ret != 0) {
        kirmes_user_record_free(records);
        return KRB5_KDB_NOENTRY;
    }

    ret = krb5_db_get_principal(kctx->kdb_base_ctx,
                                userdb_principal,
                                flags,
                                entry_out);
    if (ret == 0) {
        update_db_entry_from_user_record(context, &(records[0]), *entry_out);
    }

    kirmes_user_record_free(records);

    return ret;
}

static krb5_error_code
kuserdb_get_principal(krb5_context context,
                      krb5_const_principal searchfor,
                      unsigned int flags,
                      krb5_db_entry **entry_out)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    ret =
        krb5_db_get_principal(kctx->kdb_base_ctx, searchfor, flags, entry_out);
    if (ret == 0 || ret != KRB5_KDB_NOENTRY) {
        return ret;
    }

    if (krb5_princ_size(context, searchfor) >= 2) {
        return kuserdb_get_service_alias(context,
                                         kctx,
                                         searchfor,
                                         flags,
                                         entry_out);
    }

    return kuserdb_get_userdb_principal(context,
                                        kctx,
                                        searchfor,
                                        flags,
                                        entry_out);
}

static krb5_error_code
kuserdb_put_principal(krb5_context context,
                      krb5_db_entry *entry,
                      char **unused /* db_args */)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_put_principal(kctx->kdb_base_ctx, entry);
}

static krb5_error_code
kuserdb_delete_principal(krb5_context context, krb5_const_principal searchfor)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_delete_principal(kctx->kdb_base_ctx,
                                    (krb5_principal)searchfor);
}

static krb5_error_code
kuserdb_iterate(krb5_context context, char *match_expr,
              krb5_error_code (*func)(void *, krb5_db_entry *), void *arg,
              krb5_flags iterflags)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_iterate(kctx->kdb_base_ctx,
                           match_expr,
                           func,
                           arg,
                           iterflags);
}

static krb5_error_code
kuserdb_create_policy(krb5_context context, osa_policy_ent_t policy)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_create_policy(kctx->kdb_base_ctx, policy);
}

static krb5_error_code
kuserdb_get_policy(krb5_context context, char *name, osa_policy_ent_t *policy)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_get_policy(kctx->kdb_base_ctx, name, policy);
}

static krb5_error_code
kuserdb_put_policy(krb5_context context, osa_policy_ent_t policy)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_put_policy(kctx->kdb_base_ctx, policy);
}

static krb5_error_code
kuserdb_iter_policy(krb5_context context,
                    char *match_entry,
                    osa_adb_iter_policy_func func,
                    void *arg)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_iter_policy(kctx->kdb_base_ctx, match_entry, func, arg);
}

static krb5_error_code
kuserdb_delete_policy(krb5_context context, char *policy)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_delete_policy(kctx->kdb_base_ctx, policy);
}

static krb5_error_code
kuserdb_promote_db(krb5_context context,
                   char *unused, /* conf_section */
                   char **db_args)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_promote(kctx->kdb_base_ctx, db_args);
}

static krb5_error_code
kuserdb_check_policy_as(krb5_context context, krb5_kdc_req *request,
                      krb5_db_entry *client, krb5_db_entry *server,
                      krb5_timestamp kdc_time, const char **status,
                      krb5_pa_data ***e_data)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return ret;
    }

    return krb5_db_check_policy_as(kctx->kdb_base_ctx,
                                   request,
                                   client,
                                   server,
                                   kdc_time,
                                   status,
                                   e_data);
}

static void
kuserdb_audit_as_req(krb5_context context,
                     krb5_kdc_req *request,
                     const krb5_address *local_addr,
                     const krb5_address *remote_addr,
                     krb5_db_entry *client,
                     krb5_db_entry *server,
                     krb5_timestamp authtime,
                     krb5_error_code status)
{
    struct kuserdb_context *kctx = NULL;
    krb5_error_code ret;

    ret = krb5_db_get_context(context, (void **)&kctx);
    if (ret != 0) {
        return;
    }

    krb5_db_audit_as_req(kctx->kdb_base_ctx,
                         request,
                         local_addr,
                         remote_addr,
                         client,
                         server,
                         authtime,
                         status);
}

kdb_vftabl kdb_function_table = {
    .maj_ver = KRB5_KDB_DAL_MAJOR_VERSION,
    .min_ver = 0,
    .init_library = kuserdb_lib_init,
    .fini_library = kuserdb_lib_cleanup,
    .init_module = kuserdb_open,
    .fini_module = kuserdb_fini,
    .create = kuserdb_create,
    .destroy = kuserdb_destroy,
    .get_principal = kuserdb_get_principal,
    .put_principal = kuserdb_put_principal,
    .delete_principal = kuserdb_delete_principal,
    .iterate = kuserdb_iterate,
    .create_policy = kuserdb_create_policy,
    .get_policy = kuserdb_get_policy,
    .put_policy = kuserdb_put_policy,
    .iter_policy = kuserdb_iter_policy,
    .delete_policy = kuserdb_delete_policy,
    .promote_db = kuserdb_promote_db,
    .check_policy_as = kuserdb_check_policy_as,
    .audit_as_req = kuserdb_audit_as_req
};
